@@ -1,9 +1,9 @@
 # coding=utf-8
 #
 # This file is part of Hypothesis, which may be found at
-# https://github.com/HypothesisWorks/hypothesis-python
+# https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2018 David R. MacIver
+# Most of this work is copyright (C) 2013-2019 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -11,41 +11,66 @@
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
-# obtain one at http://mozilla.org/MPL/2.0/.
+# obtain one at https://mozilla.org/MPL/2.0/.
 #
 # END HEADER
 
-from __future__ import division, print_function, absolute_import
+from __future__ import absolute_import, division, print_function
 
 from collections import defaultdict
 
 import hypothesis.internal.conjecture.utils as cu
-from hypothesis.errors import NoExamples, NoSuchExample, Unsatisfiable, \
-    UnsatisfiedAssumption
-from hypothesis.control import assume, reject, _current_build_context
-from hypothesis._settings import note_deprecation
-from hypothesis.internal.compat import hrange, bit_length
-from hypothesis.utils.conventions import UniqueIdentifier
+from hypothesis._settings import Phase
+from hypothesis.control import _current_build_context, assume
+from hypothesis.errors import (
+    HypothesisException,
+    NoExamples,
+    NoSuchExample,
+    Unsatisfiable,
+    UnsatisfiedAssumption,
+)
+from hypothesis.internal.compat import bit_length, hrange
+from hypothesis.internal.conjecture.utils import (
+    LABEL_MASK,
+    calc_label_from_cls,
+    calc_label_from_name,
+    combine_labels,
+)
+from hypothesis.internal.coverage import check_function
 from hypothesis.internal.lazyformat import lazyformat
 from hypothesis.internal.reflection import get_pretty_function_description
-from hypothesis.internal.conjecture.utils import LABEL_MASK, \
-    combine_labels, calc_label_from_cls, calc_label_from_name
+from hypothesis.internal.validation import check_type
+from hypothesis.utils.conventions import UniqueIdentifier
 
-calculating = UniqueIdentifier('calculating')
+try:
+    from random import Random  # noqa
+    from typing import List, Callable, TypeVar, Generic, Optional  # noqa
+
+    Ex = TypeVar("Ex", covariant=True)
+    T = TypeVar("T")
+
+    from hypothesis.internal.conjecture.data import ConjectureData  # noqa
+
+except ImportError:  # pragma: no cover
+    Ex = "key"  # type: ignore
+    Generic = {Ex: object}  # type: ignore
+
+calculating = UniqueIdentifier("calculating")
 
 MAPPED_SEARCH_STRATEGY_DO_DRAW_LABEL = calc_label_from_name(
-    'another attempted draw in MappedSearchStrategy')
+    "another attempted draw in MappedSearchStrategy"
+)
 
 
 def one_of_strategies(xs):
     """Helper function for unioning multiple strategies."""
     xs = tuple(xs)
     if not xs:
-        raise ValueError('Cannot join an empty list of strategies')
+        raise ValueError("Cannot join an empty list of strategies")
     return OneOfStrategy(xs)
 
 
-class SearchStrategy(object):
+class SearchStrategy(Generic[Ex]):
     """A SearchStrategy is an object that knows how to explore data of a given
     type.
 
@@ -70,7 +95,7 @@ class SearchStrategy(object):
         The problem is that for properties that depend on each other, a naive
         calculation strategy may hit infinite recursion. Consider for example
         the property is_empty. A strategy defined as x = st.deferred(lambda x)
-        is certainly empty (in order ot draw a value from x we would have to
+        is certainly empty (in order to draw a value from x we would have to
         draw a value from x, for which we would have to draw a value from x,
         ...), but in order to calculate it the naive approach would end up
         calling x.is_empty in order to calculate x.is_empty in order to etc.
@@ -85,9 +110,9 @@ class SearchStrategy(object):
         and performance of parsing with derivatives." ACM SIGPLAN Notices 51.6
         (2016): 224-236.
         """
-        cache_key = 'cached_' + name
-        calculation = 'calc_' + name
-        force_key = 'force_' + name
+        cache_key = "cached_" + name
+        calculation = "calc_" + name
+        force_key = "force_" + name
 
         def forced_value(target):
             try:
@@ -138,13 +163,29 @@ class SearchStrategy(object):
 
                 # We track which strategies use which in the course of
                 # calculating their property value. If A ever uses B in
-                # the course of calculating its value, then whenveer the
+                # the course of calculating its value, then whenever the
                 # value of B changes we might need to update the value of
                 # A.
                 listeners = defaultdict(set)
             else:
                 needs_update = None
 
+            def recur2(strat):
+                def recur_inner(other):
+                    try:
+                        return forced_value(other)
+                    except AttributeError:
+                        pass
+                    listeners[other].add(strat)
+                    try:
+                        return mapping[other]
+                    except KeyError:
+                        needs_update.add(other)
+                        mapping[other] = default
+                        return default
+
+                return recur_inner
+
             count = 0
             seen = set()
             while needs_update:
@@ -167,20 +208,7 @@ class SearchStrategy(object):
                 to_update = needs_update
                 needs_update = set()
                 for strat in to_update:
-                    def recur(other):
-                        try:
-                            return forced_value(other)
-                        except AttributeError:
-                            pass
-                        listeners[other].add(strat)
-                        try:
-                            return mapping[other]
-                        except KeyError:
-                            needs_update.add(other)
-                            mapping[other] = default
-                            return default
-
-                    new_value = getattr(strat, calculation)(recur)
+                    new_value = getattr(strat, calculation)(recur2(strat))
                     if new_value != mapping[strat]:
                         needs_update.update(listeners[strat])
                         mapping[strat] = new_value
@@ -201,14 +229,14 @@ class SearchStrategy(object):
     # The fact that this returns False does not guarantee that a valid value
     # can be drawn - this is not intended to be perfect, and is primarily
     # intended to be an optimisation for some cases.
-    is_empty = recursive_property('is_empty', True)
+    is_empty = recursive_property("is_empty", True)
 
     # Returns True if values from this strategy can safely be reused without
     # this causing unexpected behaviour.
-    has_reusable_values = recursive_property('has_reusable_values', True)
+    has_reusable_values = recursive_property("has_reusable_values", True)
 
     # Whether this strategy is suitable for holding onto in a cache.
-    is_cacheable = recursive_property('is_cacheable', True)
+    is_cacheable = recursive_property("is_cacheable", True)
 
     def calc_is_cacheable(self, recur):
         return True
@@ -226,6 +254,7 @@ class SearchStrategy(object):
         return False
 
     def example(self, random=None):
+        # type: (Random) -> Ex
         """Provide an example of the sort of value that this strategy
         generates. This is biased to be slightly simpler than is typical for
         values from this strategy, for clarity purposes.
@@ -238,29 +267,22 @@ class SearchStrategy(object):
         context = _current_build_context.value
         if context is not None:
             if context.data is not None and context.data.depth > 0:
-                note_deprecation(
-                    'Using example() inside a strategy definition is a bad '
-                    'idea. It will become an error in a future version of '
-                    "Hypothesis, but it's unlikely that it's doing what you "
-                    'intend even now. Instead consider using '
-                    'hypothesis.strategies.builds() or '
-                    '@hypothesis.strategies.composite to define your strategy.'
-                    ' See '
-                    'https://hypothesis.readthedocs.io/en/latest/data.html'
-                    '#hypothesis.strategies.builds or '
-                    'https://hypothesis.readthedocs.io/en/latest/data.html'
-                    '#composite-strategies for more details.'
+                raise HypothesisException(
+                    "Using example() inside a strategy definition is a bad "
+                    "idea. Instead consider using hypothesis.strategies.builds() "
+                    "or @hypothesis.strategies.composite to define your strategy."
+                    " See https://hypothesis.readthedocs.io/en/latest/data.html"
+                    "#hypothesis.strategies.builds or "
+                    "https://hypothesis.readthedocs.io/en/latest/data.html"
+                    "#composite-strategies for more details."
                 )
             else:
-                note_deprecation(
-                    'Using example() inside a test function is a bad '
-                    'idea. It will become an error in a future version of '
-                    "Hypothesis, but it's unlikely that it's doing what you "
-                    'intend even now. Instead consider using '
-                    'hypothesis.strategies.data() to draw '
-                    'more examples during testing. See '
-                    'https://hypothesis.readthedocs.io/en/latest/data.html'
-                    '#drawing-interactively-in-tests for more details.'
+                raise HypothesisException(
+                    "Using example() inside a test function is a bad "
+                    "idea. Instead consider using hypothesis.strategies.data() "
+                    "to draw more examples during testing. See "
+                    "https://hypothesis.readthedocs.io/en/latest/data.html"
+                    "#drawing-interactively-in-tests for more details."
                 )
 
         from hypothesis import find, settings, Verbosity
@@ -268,7 +290,7 @@ class SearchStrategy(object):
         # Conjecture will always try the zero example first. This would result
         # in us producing the same example each time, which is boring, so we
         # deliberately skip the first example it feeds us.
-        first = []
+        first = []  # type: list
 
         def condition(x):
             if first:
@@ -276,38 +298,36 @@ class SearchStrategy(object):
             else:
                 first.append(x)
                 return False
+
         try:
             return find(
                 self,
                 condition,
                 random=random,
                 settings=settings(
-                    max_shrinks=0,
-                    max_iterations=1000,
                     database=None,
                     verbosity=Verbosity.quiet,
-                )
+                    phases=tuple(set(Phase) - {Phase.shrink}),
+                ),
             )
         except (NoSuchExample, Unsatisfiable):
             # This can happen when a strategy has only one example. e.g.
             # st.just(x). In that case we wanted the first example after all.
             if first:
                 return first[0]
-            raise NoExamples(
-                u'Could not find any valid examples in 100 tries'
-            )
+            raise NoExamples(u"Could not find any valid examples in 100 tries")
 
     def map(self, pack):
+        # type: (Callable[[Ex], T]) -> SearchStrategy[T]
         """Returns a new strategy that generates values by generating a value
         from this strategy and then calling pack() on the result, giving that.
 
         This method is part of the public API.
         """
-        return MappedSearchStrategy(
-            pack=pack, strategy=self
-        )
+        return MappedSearchStrategy(pack=pack, strategy=self)
 
     def flatmap(self, expand):
+        # type: (Callable[[Ex], SearchStrategy[T]]) -> SearchStrategy[T]
         """Returns a new strategy that generates values by generating a value
         from this strategy, say x, then generating a value from
         strategy(expand(x))
@@ -315,11 +335,11 @@ class SearchStrategy(object):
         This method is part of the public API.
         """
         from hypothesis.searchstrategy.flatmapped import FlatMapStrategy
-        return FlatMapStrategy(
-            expand=expand, strategy=self
-        )
+
+        return FlatMapStrategy(expand=expand, strategy=self)
 
     def filter(self, condition):
+        # type: (Callable[[Ex], bool]) -> SearchStrategy[Ex]
         """Returns a new strategy that generates values from this strategy
         which satisfy the provided condition. Note that if the condition is too
         hard to satisfy this might result in your tests failing with
@@ -327,13 +347,17 @@ class SearchStrategy(object):
 
         This method is part of the public API.
         """
-        return FilteredStrategy(
-            condition=condition,
-            strategy=self,
-        )
+        return FilteredStrategy(conditions=(condition,), strategy=self)
+
+    def do_filtered_draw(self, data, filter_strategy):
+        # Hook for strategies that want to override the behaviour of
+        # FilteredStrategy. Most strategies don't, so by default we delegate
+        # straight back to the default filtered-draw implementation.
+        return filter_strategy.default_do_filtered_draw(data)
 
     @property
     def branches(self):
+        # type: () -> List[SearchStrategy[Ex]]
         return [self]
 
     def __or__(self, other):
@@ -343,10 +367,11 @@ class SearchStrategy(object):
         This method is part of the public API.
         """
         if not isinstance(other, SearchStrategy):
-            raise ValueError('Cannot | a SearchStrategy with %r' % (other,))
+            raise ValueError("Cannot | a SearchStrategy with %r" % (other,))
         return one_of_strategies((self, other))
 
     def validate(self):
+        # type: () -> None
         """Throw an exception if the strategy is not valid.
 
         This can happen due to lazy construction
@@ -362,7 +387,7 @@ class SearchStrategy(object):
             self.validate_called = False
             raise
 
-    LABELS = {}
+    LABELS = {}  # type: dict
 
     @property
     def class_label(self):
@@ -391,7 +416,8 @@ class SearchStrategy(object):
         pass
 
     def do_draw(self, data):
-        raise NotImplementedError('%s.do_draw' % (type(self).__name__,))
+        # type: (ConjectureData) -> Ex
+        raise NotImplementedError("%s.do_draw" % (type(self).__name__,))
 
     def __init__(self):
         pass
@@ -413,12 +439,17 @@ class OneOfStrategy(SearchStrategy):
         self.__element_strategies = None
         self.bias = bias
         self.__in_branches = False
-        if bias is not None:
-            assert 0 < bias < 1
-            self.sampler = cu.Sampler(
-                [bias ** i for i in range(len(strategies))])
-        else:
-            self.sampler = None
+        self.__sampler = None
+
+    @property
+    def sampler(self):
+        if self.__sampler is None:
+            assert self.bias is not None
+            assert 0 < self.bias < 1
+            self.__sampler = cu.Sampler(
+                [self.bias ** i for i in range(len(self.element_strategies))]
+            )
+        return self.__sampler
 
     def calc_is_empty(self, recur):
         return all(recur(e) for e in self.original_strategies)
@@ -431,14 +462,12 @@ class OneOfStrategy(SearchStrategy):
 
     @property
     def element_strategies(self):
-        from hypothesis.strategies import check_strategy
         if self.__element_strategies is None:
             strategies = []
             for arg in self.original_strategies:
                 check_strategy(arg)
                 if not arg.is_empty:
-                    strategies.extend(
-                        [s for s in arg.branches if not s.is_empty])
+                    strategies.extend([s for s in arg.branches if not s.is_empty])
             pruned = []
             seen = set()
             for s in strategies:
@@ -452,7 +481,8 @@ class OneOfStrategy(SearchStrategy):
             shift = bit_length(len(pruned))
             for i, p in enumerate(pruned):
                 branch_labels.append(
-                    (((self.label ^ p.label) << shift) + i) & LABEL_MASK)
+                    (((self.label ^ p.label) << shift) + i) & LABEL_MASK
+                )
             self.__element_strategies = pruned
             self.__branch_labels = tuple(branch_labels)
         return self.__element_strategies
@@ -460,29 +490,31 @@ class OneOfStrategy(SearchStrategy):
     @property
     def branch_labels(self):
         self.element_strategies
+        assert len(self.__branch_labels) == len(self.element_strategies)
         return self.__branch_labels
 
     def calc_label(self):
-        return combine_labels(self.class_label, *[
-            p.label for p in self.original_strategies
-        ])
+        return combine_labels(
+            self.class_label, *[p.label for p in self.original_strategies]
+        )
 
     def do_draw(self, data):
+        # type: (ConjectureData) -> Ex
         n = len(self.element_strategies)
         assert n > 0
         if n == 1:
             return data.draw(self.element_strategies[0])
 
-        if self.sampler is None:
+        if self.bias is None:
             i = cu.integer_range(data, 0, n - 1)
         else:
             i = self.sampler.sample(data)
+            assert 0 <= i < n
 
-        return data.draw(
-            self.element_strategies[i], label=self.branch_labels[i])
+        return data.draw(self.element_strategies[i], label=self.branch_labels[i])
 
     def __repr__(self):
-        return 'one_of(%s)' % ', '.join(map(repr, self.original_strategies))
+        return "one_of(%s)" % ", ".join(map(repr, self.original_strategies))
 
     def do_validate(self):
         for e in self.element_strategies:
@@ -520,23 +552,23 @@ class MappedSearchStrategy(SearchStrategy):
         return recur(self.mapped_strategy)
 
     def __repr__(self):
-        if not hasattr(self, '_cached_repr'):
-            self._cached_repr = '%r.map(%s)' % (
-                self.mapped_strategy, get_pretty_function_description(
-                    self.pack)
+        if not hasattr(self, "_cached_repr"):
+            self._cached_repr = "%r.map(%s)" % (
+                self.mapped_strategy,
+                get_pretty_function_description(self.pack),
             )
         return self._cached_repr
 
     def do_validate(self):
         self.mapped_strategy.validate()
 
-    def pack(self, x):
+    def pack(self, x):  # type: ignore
         """Take a value produced by the underlying mapped_strategy and turn it
         into a value suitable for outputting from this strategy."""
-        raise NotImplementedError(
-            '%s.pack()' % (self.__class__.__name__))
+        raise NotImplementedError("%s.pack()" % (self.__class__.__name__))
 
     def do_draw(self, data):
+        # type: (ConjectureData) -> Ex
         for _ in range(3):
             i = data.index
             try:
@@ -548,22 +580,37 @@ class MappedSearchStrategy(SearchStrategy):
                 data.stop_example(discard=True)
                 if data.index == i:
                     raise
-        reject()
+        raise UnsatisfiedAssumption()
 
     @property
     def branches(self):
+        # type: () -> List[SearchStrategy[Ex]]
         return [
             MappedSearchStrategy(pack=self.pack, strategy=strategy)
             for strategy in self.mapped_strategy.branches
         ]
 
 
-class FilteredStrategy(SearchStrategy):
+filter_not_satisfied = UniqueIdentifier("filter not satisfied")
 
-    def __init__(self, strategy, condition):
+
+class FilteredStrategy(SearchStrategy):
+    def __init__(self, strategy, conditions):
         super(FilteredStrategy, self).__init__()
-        self.condition = condition
-        self.filtered_strategy = strategy
+        if isinstance(strategy, FilteredStrategy):
+            # Flatten chained filters into a single filter with multiple
+            # conditions.
+            self.flat_conditions = strategy.flat_conditions + conditions
+            self.filtered_strategy = strategy.filtered_strategy
+        else:
+            self.flat_conditions = conditions
+            self.filtered_strategy = strategy
+
+        assert self.flat_conditions
+        assert isinstance(self.flat_conditions, tuple)
+        assert not isinstance(self.filtered_strategy, FilteredStrategy)
+
+        self.__condition = None
 
     def calc_is_empty(self, recur):
         return recur(self.filtered_strategy)
@@ -572,17 +619,49 @@ class FilteredStrategy(SearchStrategy):
         return recur(self.filtered_strategy)
 
     def __repr__(self):
-        if not hasattr(self, '_cached_repr'):
-            self._cached_repr = '%r.filter(%s)' % (
-                self.filtered_strategy, get_pretty_function_description(
-                    self.condition)
+        if not hasattr(self, "_cached_repr"):
+            self._cached_repr = "%r%s" % (
+                self.filtered_strategy,
+                "".join(
+                    ".filter(%s)" % get_pretty_function_description(cond)
+                    for cond in self.flat_conditions
+                ),
             )
         return self._cached_repr
 
     def do_validate(self):
         self.filtered_strategy.validate()
 
+    @property
+    def condition(self):
+        if self.__condition is None:
+            assert self.flat_conditions
+            if len(self.flat_conditions) == 1:
+                # Avoid an extra indirection in the common case of only one
+                # condition.
+                self.__condition = self.flat_conditions[0]
+            else:
+                self.__condition = lambda x: all(
+                    cond(x) for cond in self.flat_conditions
+                )
+        return self.__condition
+
     def do_draw(self, data):
+        # type: (ConjectureData) -> Ex
+        result = self.filtered_strategy.do_filtered_draw(
+            data=data, filter_strategy=self
+        )
+        if result is not filter_not_satisfied:
+            return result
+
+        data.note_event("Aborted test because unable to satisfy %r" % (self,))
+        data.mark_invalid()
+        raise AssertionError("Unreachable, for Mypy")  # pragma: no cover
+
+    def note_retried(self, data):
+        data.note_event(lazyformat("Retried draw from %r to satisfy filter", self))
+
+    def default_do_filtered_draw(self, data):
         for i in hrange(3):
             start_index = data.index
             value = data.draw(self.filtered_strategy)
@@ -590,20 +669,23 @@ class FilteredStrategy(SearchStrategy):
                 return value
             else:
                 if i == 0:
-                    data.note_event(lazyformat(
-                        'Retried draw from %r to satisfy filter', self,))
+                    self.note_retried(data)
                 # This is to guard against the case where we consume no data.
                 # As long as we consume data, we'll eventually pass or raise.
                 # But if we don't this could be an infinite loop.
                 assume(data.index > start_index)
-        data.note_event('Aborted test because unable to satisfy %r' % (
-            self,
-        ))
-        data.mark_invalid()
+
+        return filter_not_satisfied
 
     @property
     def branches(self):
+        # type: () -> List[SearchStrategy[Ex]]
         return [
-            FilteredStrategy(strategy=strategy, condition=self.condition)
+            FilteredStrategy(strategy=strategy, conditions=self.flat_conditions)
             for strategy in self.filtered_strategy.branches
         ]
+
+
+@check_function
+def check_strategy(arg, name=""):
+    check_type(SearchStrategy, arg, name)
